import { Construct } from 'constructs';
import { AwsNetworkingConfig as NetworkingConfig } from './aws_config';
import { vpc } from '@cdktf/provider-aws';
import { getKeys, assert } from '../util';
import { PartialRecord } from '../base';
import { Subnets, Vpcs } from './config';
import { Direction } from './types';

export class AwsNetwork extends Construct {
  readonly config: NetworkingConfig;
  readonly vpcs: Record<Vpcs, vpc.Vpc>;
  readonly subnets: Record<Vpcs, Record<Subnets, vpc.Subnet>>;
  readonly peers: PartialRecord<
    Vpcs,
    PartialRecord<Vpcs, vpc.VpcPeeringConnection>
  >;
  readonly firewalls: Record<
    Vpcs,
    Record<Direction, Record<string, vpc.NetworkAclRule[]>>
  >;

  constructor(
    scope: Construct,
    name: string,
    config: NetworkingConfig,
  ) {
    super(scope, name);
    this.config = config;
    this.vpcs = this.networks(config);
    this.subnets = this.subnetworks(config);
    this.peers = this.networkspeering(config);
    this.firewalls = this.firewall(config);
  }

  networks(i: NetworkingConfig) {
    return getKeys(i.config)
      .map((v: Vpcs) => {
        const rv = new vpc.Vpc(this, v, {
          cidrBlock: i.config[v].cidrBlock,
        });
        return { [v]: rv } as Record<Vpcs, vpc.Vpc>;
      })
      .reduce((obj, i) => {
        return { ...obj, ...i };
      });
  }

  subnetworks(i: NetworkingConfig) {
    return getKeys(i.config)
      .map((v: Vpcs) => {
        return {
          [v]: getKeys(i.config[v].subnets)
            .map((s: Subnets) => {
              const c = assert(i.config[v].subnets[s]).config;
              return {
                [s]: new vpc.Subnet(this, `${v}-${s}`, {
                  vpcId: this.vpcs[v].id,
                  ...c,
                }),
              } as Record<Subnets, vpc.Subnet>;
            })
            .reduce((obj, i) => {
              return { ...obj, ...i };
            }),
        } as Record<Vpcs, Record<Subnets, vpc.Subnet>>;
      })
      .reduce((obj, i) => {
        return { ...obj, ...i };
      });
  }

  networkspeering(i: NetworkingConfig) {
    return getKeys(i.peers)
      .map((src) => {
        return assert(i.peers[src])
          .map((dst) => {
            return {
              [src]: {
                [dst]: new vpc.VpcPeeringConnection(
                  this,
                  `${src}-${dst}`,
                  {
                    peerVpcId: this.vpcs[dst].id,
                    vpcId: this.vpcs[src].id,
                  },
                ),
              },
            } as Record<Vpcs, Record<Vpcs, vpc.VpcPeeringConnection>>;
          })
          .reduce((obj, i) => {
            return { ...obj, ...i };
          });
      })
      .reduce((obj, i) => {
        return { ...obj, ...i };
      });
  }

  firewall(i: NetworkingConfig) {
    return getKeys(i.firewall)
      .map((v) => {
        return getKeys(i.firewall[v])
          .map((direction) => {
            const ad = (i: boolean): 'allow' | 'deny' => {
              return i == true ? 'allow' : 'deny';
            };

            const nacl = new vpc.NetworkAcl(this, `${v}-acl-${direction}`, {
              vpcId: this.vpcs[v].id,
            }); // yeah, I should work this into the return type
            return i.firewall[v][direction]
              .map((j) => {
                const sourceRanges =
                  j.sourceRanges != undefined ? j.sourceRanges : [];

                const sourceNetworks =
                  j.sourceNetworks != undefined
                    ? getKeys(j.sourceNetworks).map((net) => {
                        return this.subnets[net][
                          // typechecker ain't that great
                          assert(assert(j.sourceNetworks)[net])
                        ].cidrBlock;
                      })
                    : [];

                return {
                  [v]: {
                    [direction]: {
                      [j.name]: [
                        ...sourceRanges,
                        ...sourceNetworks,
                      ].map((block) => {
                        return new vpc.NetworkAclRule(
                          this,
                          `${v}-${direction}-${ad(j.allow)}-${j.name}`,
                          {
                            networkAclId: nacl.id,
                            ruleAction: ad(j.allow),
                            protocol: j.protocol,
                            toPort: j.port,
                            ruleNumber: j.priority,
                            egress:
                              direction == 'ingress' ? false : true,
                            cidrBlock: block,
                          },
                        );
                      }),
                    },
                  },
                } as Record<
                  Vpcs,
                  Record<
                    Direction,
                    Record<string, vpc.NetworkAclRule[]>
                  >
                >;
              })
              .reduce((obj, i) => {
                return { ...obj, ...i };
              });
          })
          .reduce((obj, i) => {
            return { ...obj, ...i };
          });
      })
      .reduce((obj, i) => {
        return { ...obj, ...i };
      });
  }
}
