import { Construct } from 'constructs';
import { GcpNetworkingConfig as NetworkingConfig } from './gcp_config';

import { Subnets, Vpcs } from './config';
import { Direction } from './types';
import {
  ComputeNetwork,
  ComputeSubnetwork,
  ComputeFirewall,
  ComputeNetworkPeering,
} from '@cdktf/provider-google';
import { getKeys, assert } from '../util';
import { PartialRecord } from '../base';

export class GcpNetwork extends Construct {
  readonly config: NetworkingConfig;
  readonly vpcs: Record<Vpcs, ComputeNetwork>;
  readonly subnets: Record<Vpcs, Record<Subnets, ComputeSubnetwork>>;
  readonly peers: PartialRecord<
    Vpcs,
    PartialRecord<Vpcs, ComputeNetworkPeering>
  >;
  readonly firewalls: Record<
    Vpcs,
    Record<Direction, Record<string, ComputeFirewall>>
  >;
  constructor(
    scope: Construct,
    name: string,
    config: NetworkingConfig,
  ) {
    super(scope, name);
    this.config = config;
    this.vpcs = this.networks(config);
    this.subnets = this.subnetworks(config);
    this.peers = this.networkspeering(config);
    this.firewalls = this.firewall(config);
  }

  networks(i: NetworkingConfig) {
    return getKeys(i.config)
      .map((v: Vpcs) => {
        const rv = new ComputeNetwork(this, v, {
          name: v,
          project: i.config[v].project,
          autoCreateSubnetworks: false,
          deleteDefaultRoutesOnCreate: true,
        });
        return { [v]: rv } as Record<Vpcs, ComputeNetwork>;
      })
      .reduce((obj, i) => {
        return { ...obj, ...i };
      });
  }

  subnetworks(i: NetworkingConfig) {
    return getKeys(i.config)
      .map((v: Vpcs) => {
        return {
          [v]: getKeys(i.config[v].subnets)
            .map((s: Subnets) => {
              const c = assert(i.config[v].subnets[s]).config;
              return {
                [s]: new ComputeSubnetwork(this, `${v}-${s}`, {
                  ...c,
                  name: `${v}-${s}`,
                  network: this.vpcs[v].name,
                }),
              } as Record<Subnets, ComputeSubnetwork>;
            })
            .reduce((obj, i) => {
              return { ...obj, ...i };
            }),
        } as Record<Vpcs, Record<Subnets, ComputeSubnetwork>>;
      })
      .reduce((obj, i) => {
        return { ...obj, ...i };
      });
  }

  networkspeering(i: NetworkingConfig) {
    return getKeys(i.peers)
      .map((src) => {
        return assert(i.peers[src])
          .map((dst) => {
            return {
              [src]: {
                [dst]: new ComputeNetworkPeering(
                  this,
                  `${src}-${dst}`,
                  {
                    name: `${src}-${dst}`,
                    network: this.vpcs[src].id,
                    peerNetwork: this.vpcs[dst].id,
                  },
                ),
              },
            } as Record<Vpcs, Record<Vpcs, ComputeNetworkPeering>>;
          })
          .reduce((obj, i) => {
            return { ...obj, ...i };
          });
      })
      .reduce((obj, i) => {
        return { ...obj, ...i };
      });
  }

  firewall(i: NetworkingConfig) {
    return getKeys(i.firewall)
      .map((v) => {
        return getKeys(i.firewall[v])
          .map((direction) => {
            const ad = (i: boolean): 'allow' | 'deny' => {
              return i == true ? 'allow' : 'deny';
            };
            return i.firewall[v][direction]
              .map((j) => {
                const sourceRanges =
                  j.sourceRanges != undefined ? j.sourceRanges : [];

                const sourceNetworks =
                  j.sourceNetworks != undefined
                    ? getKeys(j.sourceNetworks).map((net) => {
                        return this.subnets[net][
                          // typechecker ain't that great
                          assert(assert(j.sourceNetworks)[net])
                        ].ipCidrRange;
                      })
                    : [];
                return {
                  [v]: {
                    [direction]: {
                      [j.name]: new ComputeFirewall(
                        this,
                        `${v}-${direction}-${ad(j.allow)}-${j.name}`,
                        {
                          name: `${v}-${direction}-${j.name}`,
                          project: this.vpcs[v].project,
                          [ad(j.allow)]: {
                            protocol: j.protocol,
                            ports: j.port,
                          },
                          priority: j.priority,
                          sourceTags: j.sourceTags,
                          sourceRanges: [
                            ...sourceRanges,
                            ...sourceNetworks,
                          ],
                          network: this.vpcs[v].id,
                          direction: direction.toUpperCase(),
                          logConfig: {
                            metadata: 'INCLUDE_ALL_METADATA',
                          },
                        },
                      ),
                    },
                  },
                } as Record<
                  Vpcs,
                  Record<Direction, Record<string, ComputeFirewall>>
                >;
              })
              .reduce((obj, i) => {
                return { ...obj, ...i };
              });
          })
          .reduce((obj, i) => {
            return { ...obj, ...i };
          });
      })
      .reduce((obj, i) => {
        return { ...obj, ...i };
      });
  }
}
